from django.conf import settings
from django.core.cache import caches


# Force the user to create a separate cache
cache = caches[getattr(settings, 'TOS_CACHE_NAME', 'default')]


def invalidate_cached_agreements(sender, **kwargs):
    if kwargs.get('raw', False):
        return

    # Set the key version to 0 if it doesn't exist and leave it
    # alone if it does
    cache.add('django:tos:key_version', 0)

    # This key will be used to version the rest of the TOS keys
    # Incrementing it will effectively invalidate all previous keys
    cache.incr('django:tos:key_version')



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import pre_save

from .signal_handlers import invalidate_cached_agreements


class TOSConfig(AppConfig):
    name = 'tos'
    verbose_name = 'Terms Of Service'

    def ready(self):
        MIDDLEWARES = getattr(settings, 'MIDDLEWARE', [])
        if 'tos.middleware.UserAgreementMiddleware' in MIDDLEWARES:
            TermsOfService = self.get_model('TermsOfService')

            pre_save.connect(invalidate_cached_agreements,
                             sender=TermsOfService,
                             dispatch_uid='invalidate_cached_agreements')

            # Create the TOS key version immediately
            invalidate_cached_agreements(TermsOfService)



from django.urls import re_path

from tos.views import check_tos, TosView


urlpatterns = [
    # Terms of Service conform
    re_path(r'^confirm/$', check_tos, name='tos_check_tos'),

    # Terms of service simple display
    re_path(r'^$', TosView.as_view(), name='tos'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-19 14:09
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='TermsOfService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('active', models.BooleanField(default=False, help_text='Only one terms of service is allowed to be active', verbose_name='active')),
                ('content', models.TextField(blank=True, verbose_name='content')),
            ],
            options={
                'ordering': ('-created',),
                'get_latest_by': 'created',
                'verbose_name': 'Terms of Service',
                'verbose_name_plural': 'Terms of Service',
            },
        ),
        migrations.CreateModel(
            name='UserAgreement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('terms_of_service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='terms', to='tos.TermsOfService')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_agreement', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.urls import include, re_path
from django.views.generic import TemplateView

from tos import views


urlpatterns = [
    re_path(r'^$', TemplateView.as_view(template_name='index.html'), name='index'),

    re_path(r'^login/$', views.login, {}, 'login'),
    re_path(r'^tos/', include('tos.urls')),
]



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME, get_user_model
from django.core.cache import caches
from django.test import TestCase
from django.test.utils import modify_settings
from django.urls import reverse

from tos.middleware import UserAgreementMiddleware
from tos.models import TermsOfService, UserAgreement
from tos.signal_handlers import invalidate_cached_agreements


@modify_settings(
    MIDDLEWARE={
        'append': 'tos.middleware.UserAgreementMiddleware',
    },
)
class TestMiddleware(TestCase):

    def setUp(self):
        # Clear cache between tests
        cache = caches[getattr(settings, 'TOS_CACHE_NAME', 'default')]
        cache.clear()

        # User that has agreed to TOS
        self.user1 = get_user_model().objects.create_user('user1', 'user1@example.com', 'user1pass')

        # User that has not yet agreed to TOS
        self.user2 = get_user_model().objects.create_user('user2', 'user2@example.com', 'user2pass')
        self.user3 = get_user_model().objects.create_user('user3', 'user3@example.com', 'user3pass')

        self.tos1 = TermsOfService.objects.create(
            content="first edition of the terms of service",
            active=True
        )
        self.tos2 = TermsOfService.objects.create(
            content="second edition of the terms of service",
            active=False
        )
        self.login_url = getattr(settings, 'LOGIN_URL', '/login/')

        UserAgreement.objects.create(
            terms_of_service=self.tos1,
            user=self.user1
        )

        self.redirect_page = '{0}?{1}={2}'.format(
            reverse('tos_check_tos'),
            REDIRECT_FIELD_NAME,
            reverse('index'),
        )

    def test_middleware_redirects(self):
        """
        User that hasn't accepted TOS should be redirected to confirm. Also make sure
        confirm works.
        """
        self.client.login(username='user2', password='user2pass')
        response = self.client.get(reverse('index'))
        self.assertRedirects(response, self.redirect_page)

        # Make sure confirm works after middleware redirect.
        response = self.client.post(reverse('tos_check_tos'), {'accept': 'accept'})

        # Confirm redirects.
        self.assertEqual(response.status_code, 302)

    def test_invalidate_cache_on_accept_fix_redirect_loop(self):
        """
        Make sure accepting doesnt send you right back to tos page.
        """
        self.assertFalse(UserAgreement.objects.filter(terms_of_service=self.tos1, user=self.user2).exists())

        self.client.login(username='user2', password='user2pass')
        response = self.client.get(reverse('index'))
        self.assertRedirects(response, self.redirect_page)

        # Make sure confirm works after middleware redirect.
        response = self.client.post(reverse('tos_check_tos'), {'accept': 'accept'})

        self.assertTrue(UserAgreement.objects.filter(terms_of_service=self.tos1, user=self.user2).exists())

        response = self.client.get(reverse('index'))
        self.assertEqual(response.status_code, 200)

        self.assertIn('index', str(response.content))

    def test_middleware_doesnt_redirect(self):
        """User that has accepted TOS should get 200."""
        self.client.login(username='user1', password='user1pass')
        response = self.client.get(reverse('index'))
        self.assertEqual(response.status_code, 200)

    def test_anonymous_user_200(self):
        response = self.client.get(reverse('index'))
        self.assertEqual(response.status_code, 200)

    def test_accept_after_middleware_redirects_properly(self):
        self.client.login(username='user3', password='user3pass')

        response = self.client.get(reverse('index'), follow=True)

        self.assertRedirects(response, self.redirect_page)

        # Agree
        response = self.client.post(self.redirect_page, {'accept': 'accept'})

        # Confirm redirects back to the index page
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url.replace('http://testserver', ''), str(reverse('index')))


@modify_settings(
    MIDDLEWARE={
        'append': 'tos.middleware.UserAgreementMiddleware',
    },
)
class BumpCoverage(TestCase):

    def setUp(self):
        # User that has agreed to TOS
        self.user1 = get_user_model().objects.create_user('user1', 'user1@example.com', 'user1pass')

        self.tos1 = TermsOfService.objects.create(
            content="first edition of the terms of service",
            active=True
        )
        self.tos2 = TermsOfService.objects.create(
            content="second edition of the terms of service",
            active=False
        )
        self.login_url = getattr(settings, 'LOGIN_URL', '/login/')

        UserAgreement.objects.create(
            terms_of_service=self.tos1,
            user=self.user1
        )

    def test_ajax_request(self):
        class Request(object):
            method = 'GET'
            META = {
                'HTTP_X_REQUESTED_WITH': 'XMLHttpRequest'
            }

            def is_ajax(self):
                return True

        mw = UserAgreementMiddleware()

        response = mw.process_request(Request())

        self.assertIsNone(response)

    def test_skip_for_user(self):
        cache = caches[getattr(settings, 'TOS_CACHE_NAME', 'default')]

        key_version = cache.get('django:tos:key_version')

        cache.set(f'django:tos:skip_tos_check:{self.user1.id}', True, version=key_version)

        self.client.login(username='user1', password='user1pass')
        response = self.client.get(reverse('index'))

        self.assertEqual(response.status_code, 200)

    def test_invalidate_cached_agreements(self):
        cache = caches[getattr(settings, 'TOS_CACHE_NAME', 'default')]

        invalidate_cached_agreements(TermsOfService)

        key_version = cache.get('django:tos:key_version')

        invalidate_cached_agreements(TermsOfService)

        self.assertEqual(cache.get('django:tos:key_version'), key_version+1)

        invalidate_cached_agreements(TermsOfService, raw=True)

        self.assertEqual(cache.get('django:tos:key_version'), key_version+1)



from modeltranslation.translator import translator, TranslationOptions
from tos.models import TermsOfService

# Translations for django-tos

class TermsOfServiceTranslationOptions(TranslationOptions):
    fields = ('content', )

translator.register(TermsOfService, TermsOfServiceTranslationOptions)



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _


class NoActiveTermsOfService(ValidationError):
    pass


class BaseModel(models.Model):
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True


class TermsOfServiceManager(models.Manager):
    def get_current_tos(self):
        try:
            return self.get(active=True)
        except self.model.DoesNotExist:
            raise NoActiveTermsOfService(
                'Please create an active Terms-of-Service'
            )


class TermsOfService(BaseModel):
    active = models.BooleanField(
                default=False,
                verbose_name=_('active'),
                help_text=_(
                    'Only one terms of service is allowed to be active'
                )
    )
    content = models.TextField(verbose_name=_('content'), blank=True)
    objects = TermsOfServiceManager()

    class Meta:
        get_latest_by = 'created'
        ordering = ('-created',)
        verbose_name = _('Terms of Service')
        verbose_name_plural = _('Terms of Service')

    def __str__(self):
        active = 'inactive'
        if self.active:
            active = 'active'
        return f'{self.created}: {active}'

    def save(self, *args, **kwargs):
        """ Ensure we're being saved properly """

        if self.active:
            TermsOfService.objects.exclude(id=self.id).update(active=False)

        else:
            if not TermsOfService.objects\
                    .exclude(id=self.id)\
                    .filter(active=True)\
                    .exists():
                raise NoActiveTermsOfService(
                    'One of the terms of service must be marked active'
                )

        super(TermsOfService, self).save(*args, **kwargs)


class UserAgreement(BaseModel):
    terms_of_service = models.ForeignKey(TermsOfService, related_name='terms', on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='user_agreement', on_delete=models.CASCADE)

    def __str__(self):
        return f'{self.user.username} agreed to TOS: {self.terms_of_service}'


def has_user_agreed_latest_tos(user):
    return UserAgreement.objects.filter(
        terms_of_service=TermsOfService.objects.get_current_tos(),
        user=user,
    ).exists()



from django.conf import settings
from django.contrib.auth import SESSION_KEY as session_key
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.cache import caches
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy
from django.utils.cache import add_never_cache_headers
from django.utils.deprecation import MiddlewareMixin

from .models import UserAgreement


cache = caches[getattr(settings, 'TOS_CACHE_NAME', 'default')]
tos_check_url = reverse_lazy('tos_check_tos')


class UserAgreementMiddleware(MiddlewareMixin):
    """
    Some middleware to check if users have agreed to the latest TOS
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def process_request(self, request):
        if self.should_fast_skip(request):
            return None

        # Grab the user ID from the session so we avoid hitting the database
        # for the user object.
        # NOTE: We use the user ID because it's not user-settable and it won't
        #       ever change (usernames and email addresses can change)
        user_id = request.session['_auth_user_id']

        # Get the cache prefix
        key_version = cache.get('django:tos:key_version')

        # Skip if the user is allowed to skip - for instance, if the user is an
        # admin or a staff member
        if cache.get(f'django:tos:skip_tos_check:{user_id}', False, version=key_version):
            return None

        # Ping the cache for the user agreement
        user_agreed = cache.get(f'django:tos:agreed:{user_id}', None, version=key_version)

        # If the cache is missing this user
        if user_agreed is None:
            # Check the database and cache the result
            user_agreed = self.get_and_cache_agreement_from_db(user_id, key_version)

        if not user_agreed:
            # Confirm view uses these session keys. Non-middleware flow sets them in login view,
            # so we need to set them here.
            request.session['tos_user'] = user_id
            request.session['tos_backend'] = request.session['_auth_user_backend']

            response = HttpResponseRedirect('{0}?{1}={2}'.format(
                tos_check_url,
                REDIRECT_FIELD_NAME,
                request.path_info,
            ))
            add_never_cache_headers(response)
            return response

        return None

    def should_fast_skip(self, request):
        '''Check if we should skip TOS checks without hitting the cache or database'''
        # Don't get in the way of any mutating requests
        if request.method != 'GET':
            return True

        # Ignore ajax requests
        if request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest':
            return True

        # Don't redirect users when they're trying to get to the confirm page
        if request.path_info == tos_check_url:
            return True

        # If the user doesn't have a user ID, ignore them - they're anonymous
        if not request.session.get(session_key, None):
            return True

        return False

    def get_and_cache_agreement_from_db(self, user_id, key_version):
        # Grab the data from the database
        user_agreed = UserAgreement.objects.filter(
            user__id=user_id,
            terms_of_service__active=True).exists()

        # Set the value in the cache
        cache.set(f'django:tos:agreed:{user_id}', user_agreed, version=key_version)

        return user_agreed
